#!/usr/bin/python
"""
This script attempts to make it easy to run twisted in three different modes to
test performance while serving some static content.

Provide one of the following command line arguments to specify:
  twisted  -  Run using twisted only
  oldqx    -  Run using twisted and quixote, using the standard
              facilities in quixote.server.twisted_http
  newqx    -  Run using twisted and quixote, using the facilities 
              in this file.

Then fire up a client (Moz or I.E., for example) and point it to the index.html
of the running web server, e.g. http://localhost:%(WEB_PORT)s/index.html.

The index.html page consists of a 10x10 table with an image (a red square) in
each cell.  There are 100 distinct images, so that the browser will have to
make 100 (plus 1 for the page itself) distinct requests to the server to render
the page.

When doing reloads to see the time taken to render them, be sure to force a
full reload (in Moz on Windows, you can do this by holding shift as you use the
mouse to click the reload button in the toolbar, or by hitting ctrl-f5.  In
I.E, ctrl-f5 works, or a ctrl-reload click. In Moz on linux, shift-reload
works, and ctrl-shft-r works.)


On my system, pure twisted outperforms the other two, but in one case (newqx)
it's only by a little bit, but in the other case (oldqx) it's by a HUGE margin
(on the order of twenty times faster.)

IMPORTANT: THE PERFORMANCE DIFFERENCES ARE ONLY PRESENT WHEN USING HTTP
KEEP-ALIVES.  If you're using a client that doesn't support Keep-Alives, you
won't notice a difference between the three.

It's counter intuitive, but on my system, I actually get much better
performance out of the oldqx configuration when keep-alives are disabled.  This
will actually slow the pure twisted configuration down a little bit, but not
too noticably 'seat of the pants'.

Note:  Twisted seems to support pipelining, but I haven't yet figured out how
to get Quixote to play nicely enough to make it work when Quixote is in the
picture.  (But turn pipelining on in your browser, and run the test page using
twisted mode, and watch it fly!)
"""

from twisted.web import server
from twisted.protocols import http
import urllib
from quixote.http_response import Stream

class QuixoteTWRequest(server.Request):
    def process(self):
        self.publisher = self.channel.factory.publisher
        environ = self.create_environment()
        ## this seek is important, it doesnt work without it
        self.content.seek(0,0)
        qxrequest = self.publisher.create_request(self.content, environ)
        self.quixote_publish(qxrequest, environ)
        resp = qxrequest.response
        self.setResponseCode(resp.status_code)
        for hdr, value in resp.generate_headers():
            self.setHeader(hdr, value)
        
        #if resp.body is not None:
        #    self.write(resp.body)
        
        ## This is cheating, but it helps...
        #self.channel.transport.loseConnection()
        #self.finish()

        Buffer(resp.body, self)        


    def quixote_publish(self, qxrequest, env):
        """
        Warning, this sidesteps the Publisher.publish method,
        Hope you didn't override it...
        """
        pub = self.publisher
        output = pub.process_request(qxrequest, env)

        # don't write out the output, just set the response body
        # the calling method will do the rest.
        if output:
            if isinstance(output, Stream):
                # XXX a Twisted person needs to add support for streams
                output = ''.join(map(str, output))
            qxrequest.response.set_body(output)

        pub._clear_request()


    def create_environment(self):
        """
        Borrowed heavily from twisted.web.twcgi
        """
        
        # Twisted doesn't decode the path for us,
        # so let's do it here.  This is also
        # what medusa_http.py does, right or wrong.
        if '%' in self.path:
            self.path = urllib.unquote(self.path)
        
        serverName = self.getRequestHostname().split(':')[0]
        env = {"SERVER_SOFTWARE":   server.version,
               "SERVER_NAME":       serverName,
               "GATEWAY_INTERFACE": "CGI/1.1",
               "SERVER_PROTOCOL":   self.clientproto,
               "SERVER_PORT":       str(self.getHost()[2]),
               "REQUEST_METHOD":    self.method,
               "SCRIPT_NAME":       '',
               "SCRIPT_FILENAME":   '',
               "REQUEST_URI":       self.uri,
               "HTTPS":             self.isSecure() and 'on' or 'off',
        }

        client = self.getClient()
        if client is not None:
            env['REMOTE_HOST'] = client
        ip = self.getClientIP()
        if ip is not None:
            env['REMOTE_ADDR'] = ip
        xx, xx, remote_port = self.transport.getPeer()
        env['REMOTE_PORT'] = remote_port
        env["PATH_INFO"] = self.path

        qindex = self.uri.find('?')
        if qindex != -1:
            env['QUERY_STRING'] = self.uri[qindex+1:]
        else:
            env['QUERY_STRING'] = ''

        # Propogate HTTP headers
        for title, header in self.getAllHeaders().items():
            envname = title.replace('-', '_').upper()
            if title not in ('content-type', 'content-length'):
                envname = "HTTP_" + envname
            env[envname] = header

        return env


class QuixoteFactory (http.HTTPFactory):
    def __init__(self, publisher):
        http.HTTPFactory.__init__(self, None)
        self.publisher = publisher

    def buildProtocol (self, addr):
        p = http.HTTPFactory.buildProtocol(self, addr)
        p.requestFactory = QuixoteTWRequest
        return p


####################################################
####################################################

from twisted.internet import abstract
class Buffer:
    """
    A class to represent the transfer of data over the network.
    (This is a modified version of twisted.web.static.FileTransfer)
    """
    request = None
    def __init__(self, data, request):
        self.request = request
        self.data = data
        self.size = len(data)
        self.written = 0
        request.registerProducer(self, 0)

    def resumeProducing(self):
        if not self.request:
            return
        chunkSize = min(abstract.FileDescriptor.bufferSize, self.size - self.written)
        data = self.data[:chunkSize]
        self.data = self.data[chunkSize:]
        if data:
            self.written += len(data)
            self.request.write(data)
        if not self.data:
            self.request.unregisterProducer()
            self.request.finish()
            self.request = None

    def pauseProducing(self):
        pass

    def stopProducing(self):
        self.data = ""
        self.request = None

####################################################
####################################################

#Run Modes:
USE_PURE_TWISTED = 1
USE_STANDARD_QUIXOTE = 2
USE_PROTOTYPE_QUIXOTE = 3

if __name__ == "__main__":
    from twisted.internet import reactor
    import sys, os.path
    
    # Find out what our path is
    import locator
    locator.getLocation()

    web_root_path = os.path.join(locator.getLocation(), 'test_web_root')

    WEB_PORT = 8082
    
    runMode = 0

    
    if 'twisted' in sys.argv:
        runMode = USE_PURE_TWISTED
    elif 'oldqx' in sys.argv:
        runMode = USE_STANDARD_QUIXOTE
    elif 'newqx' in sys.argv:
        runMode = USE_PROTOTYPE_QUIXOTE

    if runMode == USE_PROTOTYPE_QUIXOTE:
        from quixote.publish import Publisher
        from quixote.util import StaticDirectory
        publisher = Publisher(StaticDirectory(web_root_path, cache_time=6000))

        # use the factory and request defined in this file
        factory = QuixoteFactory(publisher)

        print "using prototype quixote stuff"
    elif runMode == USE_STANDARD_QUIXOTE:
        from quixote.publish import Publisher
        from quixote.util import StaticDirectory
        from quixote.server import twisted_http
        publisher = Publisher(StaticDirectory(web_root_path, cache_time=6000))

        # use the factory and request defined in quixote.server.twisted_http
        factory = twisted_http.QuixoteFactory(publisher)

        print "using standard quixote stuff"
    elif runMode == USE_PURE_TWISTED:
        from twisted.web import static, server
        root = static.File(web_root_path)
        factory = server.Site(root)
        print "pure twisted"
    else:
        print __doc__ % vars()
        sys.exit()
    
    reactor.listenTCP(WEB_PORT, factory)
    reactor.run()



